<?php

namespace Drupal\webform_hotdocs\Controller;

use Drupal\Component\Uuid\Php as UuidService;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\encrypt\EncryptServiceInterface;
use Drupal\webform\WebformThirdPartySettingsManagerInterface;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\StreamedResponse;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Proxies requests against the HotDocs API.
 *
 * @package Drupal\webform_hotdocs\Controller
 */
class HotDocsProxyController extends ControllerBase {

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The webform third party settings manager.
   *
   * @var \Drupal\webform\WebformThirdPartySettingsManagerInterface
   */
  protected $thirdPartySettingsManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The encryption service.
   *
   * @var \Drupal\encrypt\EncryptServiceInterface
   */
  protected $encryption;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The streamed response.
   *
   * @var \Symfony\Component\HttpFoundation\StreamedResponse
   */
  protected $streamedResponse;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\Php
   */
  protected $uuidService;

  /**
   * The key value store.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueStoreInterface
   */
  protected $keyValue;

  /**
   * The logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Module constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client service.
   * @param \Drupal\webform\WebformThirdPartySettingsManagerInterface $third_party_settings_manager
   *   The webform third party settings manager service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\encrypt\EncryptServiceInterface $encryption
   *   The encryption service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack service.
   * @param \Symfony\Component\HttpFoundation\StreamedResponse $streamed_response
   *   The streamed response service.
   * @param \Drupal\Component\Uuid\Php $uuid_service
   *   The UUID service.
   * @param \Drupal\Core\KeyValueStore\KeyValueFactoryInterface $key_value_factory
   *   The key value factory service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    ClientInterface $http_client,
    WebformThirdPartySettingsManagerInterface $third_party_settings_manager,
    EntityTypeManagerInterface $entity_type_manager,
    EncryptServiceInterface $encryption,
    RequestStack $request_stack,
    StreamedResponse $streamed_response,
    UuidService $uuid_service,
    KeyValueFactoryInterface $key_value_factory,
    LoggerChannelFactoryInterface $logger_factory,
  ) {
    $this->configFactory = $config_factory;
    $this->httpClient = $http_client;
    $this->thirdPartySettingsManager = $third_party_settings_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->encryption = $encryption;
    $this->requestStack = $request_stack;
    $this->streamedResponse = $streamed_response;
    $this->uuidService = $uuid_service;
    $this->keyValue = $key_value_factory->get('webform_hotdocs_documents');
    $this->logger = $logger_factory->get('webform_hotdocs');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('http_client'),
      $container->get('webform.third_party_settings_manager'),
      $container->get('entity_type.manager'),
      $container->get('encryption'),
      $container->get('request_stack'),
      new StreamedResponse(),
      new UuidService(),
      $container->get('keyvalue'),
      $container->get('logger.factory')
    );
  }

  /**
   * Retrieves access token from HotDocs.
   */
  public function getAccessToken() {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];

    $url = "https://{$tenancy_moniker}.{$api_base_url}/auth/Authorize/ServicePrincipalLogIn";

    $password = $settings['service_principal']['principal_password'];
    if ($settings['service_principal']['principal_password_encrypted']) {
      $encryption_profile = $this->entityTypeManager
        ->getStorage('encryption_profile')
        ->load($settings['service_principal']['principal_password_encryption_profile']);
      $password = $this->encryption->decrypt($password, $encryption_profile);
    }

    if (!$password) {
      throw new \Exception('No password found for service principal.');
    }

    $body = [
      'ClientName' => $settings['service_principal']['client_name'],
      'PrincipalName' => $settings['service_principal']['principal_name'],
      'PrincipalPassword' => $password,
    ];

    try {
      $response = $this->httpClient->request('POST', $url, [
        'json' => $body,
      ]);

      $access_token = trim($response->getBody()->getContents(), '"');
      return new JsonResponse($access_token);
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Retrieves a template package from HotDocs.
   *
   * @param string $template_package_id
   *   The template package ID.
   */
  public function getTemplatePackage($template_package_id) {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];
    // Hardcoded API version; HotDocs's v2 API does not support this endpoint.
    $api_version = '1.1';

    $url = "https://{$tenancy_moniker}.{$api_base_url}/Api/rest/v{$api_version}/TemplatePackages/{$template_package_id}";

    // Get the access token from query parameter.
    $access_token = $this->getCleanToken();

    if (empty($access_token)) {
      throw new \Exception('No access token found.');
    }

    try {
      $response = $this->httpClient->request('GET', $url, [
        'headers' => [
          'Authorization' => 'Bearer ' . $access_token,
        ],
      ]);
      $content = json_decode($response->getBody()->getContents(), TRUE);
      return new JsonResponse($content);
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Retrieves a HotDocs template package's variable data.
   *
   * @param string $template_package_version_id
   *   The template package version ID.
   */
  public function getTemplatePackageVariableData($template_package_version_id) {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];
    // Hardcoded API version; HotDocs's v2 API does not support this endpoint.
    $api_version = '1.1';

    $url = "https://{$tenancy_moniker}.{$api_base_url}/Api/rest/v{$api_version}/TemplatePackageVersions/{$template_package_version_id}/VariableData";

    // Get the access token from query parameter.
    $access_token = $this->getCleanToken();

    if (empty($access_token)) {
      throw new \Exception('No access token found.');
    }

    try {
      $response = $this->httpClient->request('GET', $url, [
        'headers' => [
          'Authorization' => 'Bearer ' . $access_token,
        ],
      ]);
      $content = json_decode($response->getBody()->getContents(), TRUE);
      return new JsonResponse($content);
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Creates a work item in HotDocs.
   *
   * @param string $guid
   *   The GUID.
   */
  public function createWorkItem($guid) {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];
    $api_version = $settings['api_version'];

    $url = "https://{$tenancy_moniker}.{$api_base_url}/Api/rest/v{$api_version}/WorkItems/{$guid}";

    // Get the access token from query parameter.
    $access_token = $this->getCleanToken();

    if (empty($access_token)) {
      throw new \Exception('No access token found.');
    }

    // Get the request body content.
    $raw_content = $this->requestStack->getCurrentRequest()->getContent();

    try {
      $response = $this->httpClient->request('PUT', $url, [
        'headers' => [
          'Authorization' => 'Bearer ' . $access_token,
          'Content-Type' => 'application/json',
        ],
        'body' => $raw_content,
      ]);
      return new JsonResponse($response->getBody()->getContents());
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Create and assemble a new version of a work item in HotDocs.
   *
   * @param string $template_package_id
   *   The template package ID.
   * @param string $work_item_id
   *   The work item ID.
   */
  public function createAndAssembleWorkItem($template_package_id, $work_item_id) {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];
    $api_version = $settings['api_version'];

    $guid = $this->uuidService->generate();
    $url = "https://{$tenancy_moniker}.{$api_base_url}/Api/rest/v{$api_version}/WorkItems/{$work_item_id}/Versions/{$guid}/CreateAndAssemble";

    // Get the access token from query parameters.
    $access_token = $this->getCleanToken();

    if (empty($access_token)) {
      throw new \Exception('No access token found.');
    }

    $body = [
      'templatePackageId' => $template_package_id,
    ];

    try {
      $response = $this->httpClient->request('PUT', $url, [
        'headers' => [
          'Authorization' => 'Bearer ' . $access_token,
          'Content-Type' => 'application/json; charset=UTF-8',
        ],
        'json' => $body,
      ]);
      return new JsonResponse($response->getBody()->getContents());
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Downloads assembled documents from HotDocs.
   *
   * @param string $work_item_id
   *   The work item ID.
   */
  public function downloadAssembledDocuments($work_item_id) {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];
    $api_version = $settings['api_version'];

    $url = "https://{$tenancy_moniker}.{$api_base_url}/Api/rest/v{$api_version}/WorkItems/{$work_item_id}/Documents";

    // Get the access token from query parameters.
    $access_token = $this->getCleanToken();

    if (empty($access_token)) {
      throw new \Exception('No access token found.');
    }

    try {
      $response = $this->httpClient->request('GET', $url, [
        'headers' => [
          'Authorization' => 'Bearer ' . $access_token,
        ],
      ]);
      return new JsonResponse($response->getBody()->getContents());
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Gets the access token from the query parameter.
   *
   * @return string
   *   The access token.
   */
  protected function getCleanToken() {
    $access_token = $this->requestStack->getCurrentRequest()->query->get('access_token');
    $access_token = trim($access_token, '"');

    if (empty($access_token)) {
      throw new \Exception('No access token found.');
    }

    return $access_token;
  }

  /**
   * Downloads the content of an assembled document from HotDocs.
   *
   * @param string $work_item_id
   *   The work item ID.
   * @param string $document_id
   *   The document ID.
   */
  public function downloadAssembledDocumentContent($work_item_id, $document_id) {
    // Load Webform third-party settings.
    $settings = $this->thirdPartySettingsManager->getThirdPartySettings('hotdocs');
    $tenancy_moniker = $settings['tenancy_moniker'];
    $api_base_url = $settings['api_base_domain'];
    $api_version = $settings['api_version'];

    $url = "https://{$tenancy_moniker}.{$api_base_url}/Api/rest/v{$api_version}/WorkItems/{$work_item_id}/Documents/{$document_id}/Content";

    try {
      $access_token = $this->getCleanToken();

      $response = $this->httpClient->request('GET', $url, [
        'headers' => [
          'Authorization' => 'Bearer ' . $access_token,
        ],
        'stream' => TRUE,
      ]);

      // Get content type from response.
      $content_type = $response->getHeader('Content-Type')[0] ?? 'application/octet-stream';

      // Create a Symfony response with the streamed content.
      $this->streamedResponse->setCallback(function () use ($response) {
        $stream = $response->getBody();
        while (!$stream->eof()) {
          $chunk = $stream->read(1024);
          echo $chunk;
        }
      });

      $this->streamedResponse->headers->set('Content-Type', $content_type);
      $this->streamedResponse->headers->set('Content-Disposition', 'inline');

      return $this->streamedResponse;
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Downloads a stored document.
   *
   * @param string $id
   *   The document ID.
   */
  public function downloadStoredDocument($id) {
    try {
      $document_data = $this->keyValue->get($id);
      if (!$document_data) {
        throw new NotFoundHttpException('Document not found.');
      }

      $base64_content = $document_data['content'];
      if (empty($base64_content)) {
        throw new \Exception('No document content found.');
      }

      $decoded = base64_decode($base64_content);
      if ($decoded === FALSE) {
        throw new \Exception('Failed to decode document content.');
      }

      // Create temporary file.
      $temp_file = tempnam(sys_get_temp_dir(), 'hotdocs_');
      if (file_put_contents($temp_file, $decoded) === FALSE) {
        throw new \Exception('Failed to write document to temporary file.');
      }

      $response = new BinaryFileResponse($temp_file);

      // Lowercase the document type to clean up HotDocs's response.
      $type = $document_data['type'] ? strtolower($document_data['type']) : 'docx';
      $name = $document_data['name'] ?? 'document';

      switch ($type) {
        case 'pdf':
          $response->setContentDisposition('attachment', $name . '.pdf');
          $response->headers->set('Content-Type', 'application/pdf');
          break;

        case 'docx':
        default:
          $response->setContentDisposition('attachment', $name . '.docx');
          $response->headers->set('Content-Type', 'application/vnd.openxmlformats-officedocument.wordprocessingml.document');
          break;
      }

      $response->deleteFileAfterSend(TRUE);

      return $response;
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to download document: @error', ['@error' => $e->getMessage()]);
      throw new NotFoundHttpException('Document not found or invalid.');
    }
  }

}
