<?php

namespace Drupal\webform_hotdocs\Plugin\WebformHandler;

use Drupal\Core\Form\FormStateInterface;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformSubmissionInterface;
use GuzzleHttp\Exception\RequestException;

/**
 * HotDocs webform handler.
 *
 * @WebformHandler(
 *   id = "hotdocs",
 *   label = @Translation("HotDocs"),
 *   category = @Translation("HotDocs"),
 *   description = @Translation("Sends data from Webform submissions to HotDocs."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 * )
 */
class HotDocsWebformHandler extends WebformHandlerBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'template_package_id' => '',
      'field_mappings' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['token_instructions'] = [
      '#type' => 'webform_message',
      '#message_type' => 'info',
      '#message_message' => $this->getHandlerId()
        ? $this->t(
          '<ol>
            <li>Add a Value field to the Webform with the following token, based on the machine name of this handler, as its value: <code>[webform:handler:@handler_id:completed:document_id]</code>. This is needed to store the ID of the HotDocs-generated document.</li>
            <li>To include a link to download the document on the confirmation page or in another handler, build a link to <code>/download-document/</code> followed by a Webform submission data token that uses the machine name of that Value field. For example: <pre>&lt;a href="/download-document/[webform_submission:values:MACHINE_NAME]" class="button"&gt;Download document&lt;/a&gt;</pre></li>
          </ol>',
          ['@handler_id' => $this->getHandlerId()]
      )
        : $this->t('Save, then re-edit this handler for instructions on how to use the generated document.'),
    ];
    $form['template_package_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Template package ID'),
      '#description' => $this->t('The HotDocs template package ID to use.'),
      '#required' => TRUE,
      '#default_value' => $this->configuration['template_package_id'],
      '#ajax' => [
        'callback' => [$this, 'templatePackageIdAjaxCallback'],
        'event' => 'change',
        'wrapper' => 'edit-field-mappings',
        'progress' => [
          'type' => 'throbber',
          'message' => $this->t('Authenticating with HotDocs...'),
        ],
      ],
    ];

    $form['field_mappings'] = [
      '#type' => 'details',
      '#title' => $this->t('Field mappings'),
      '#description' => $this->t('Map Webform fields to HotDocs fields.'),
      '#open' => TRUE,
      '#attributes' => ['id' => 'edit-field-mappings'],
    ];

    // If a template package ID is set, populate variables and field mappings.
    if (!empty($this->configuration['template_package_id'])) {
      try {
        $access_token = $this->getAccessToken();
        if ($access_token) {
          $template_package = $this->getTemplatePackage($access_token, $this->configuration['template_package_id']);
          $version_id = $template_package['liveVersionId'];
          $variableData = $this->getTemplatePackageVariableData($access_token, $version_id);
          $variables = $variableData['variables'];

          $form['field_mappings']['variables'] = [
            '#type' => 'details',
            '#title' => $this->t('Document variables'),
            '#description' => $this->t('These are the document variables for the selected template package.'),
            '#open' => FALSE,
            'table' => [
              '#type' => 'table',
              '#header' => [
                $this->t('Name'),
                $this->t('Type'),
              ],
              '#rows' => array_map(function ($variable) {
                return [
                  $variable['name'],
                  $variable['type'],
                ];
              }, $variables),
            ],
          ];

          // Prepare variables for later use.
          $variable_mappings = [];
          foreach ($variables as $variable) {
            $variable_mappings[$variable['name']] = [
              'name' => $variable['name'],
              'type' => $variable['type'],
            ];
          }

          $form['hotdocs_variables'] = [
            '#type' => 'value',
            '#value' => $variable_mappings,
          ];

          // Create options array from variables.
          $variable_options = [
            '' => $this->t('- None -'),
          ];
          foreach ($variables as $variable) {
            $variable_options[$variable['name']] = $variable['name'] . ' (' . $variable['type'] . ')';
          }

          // Get all fields from the webform.
          $webform = $this->getWebform();
          $elements = $webform->getElementsInitializedAndFlattened();

          foreach ($elements as $key => $element_info) {
            if ($this->isValidFormElement($element_info)) {
              $form['field_mappings'][$key] = [
                '#type' => 'select',
                '#title' => $element_info['#title'] ? $element_info['#title'] . ' <small>(<code>' . $key . '</code>)</small>' : $key,
                '#options' => $variable_options,
                '#default_value' => $this->configuration['field_mappings'][$key] ?? '',
              ];
            }
          }
        }
      }
      catch (\Exception $e) {
        \Drupal::logger('webform_hotdocs')->error('Failed to load HotDocs variables: @error', ['@error' => $e->getMessage()]);
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    $this->configuration['template_package_id'] = $form_state->getValue('template_package_id');
    $field_mappings = $form_state->getValue('field_mappings') ?? [];
    $this->configuration['hotdocs_variables'] = $form_state->getValue('hotdocs_variables') ?? [];
    // Only save the actual field mappings, not the variables table.
    unset($field_mappings['variables']);
    $this->configuration['field_mappings'] = $field_mappings;
  }

  /**
   * Gets an access token from HotDocs.
   *
   * @return string|null
   *   The access token if successful, null otherwise.
   *
   * @throws \Exception
   *   If the request fails.
   */
  protected function getAccessToken() {
    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.access_token', [], ['absolute' => TRUE]);
    $client = \Drupal::httpClient();

    // Set maximum number of retries and timeouts.
    $max_retries = 3;
    $attempt = 0;

    while ($attempt < $max_retries) {
      try {
        \Drupal::logger('webform_hotdocs')->notice('Attempting to get access token (attempt @attempt of @max)', [
          '@attempt' => $attempt + 1,
          '@max' => $max_retries,
        ]);

        $response = $client->post($url, [
          'timeout' => 15,
          'connect_timeout' => 5,
        ]);

        return trim((string) $response->getBody(), '"');
      }
      catch (\Exception $e) {
        $attempt++;
        \Drupal::logger('webform_hotdocs')->warning('Access token attempt @attempt failed: @error', [
          '@attempt' => $attempt,
          '@error' => $e->getMessage(),
        ]);

        if ($attempt >= $max_retries) {
          throw new \Exception('Failed to get access token after ' . $max_retries . ' attempts: ' . $e->getMessage());
        }

        // Wait before retrying (exponential backoff).
        sleep(pow(2, $attempt - 1));
      }
    }
  }

  /**
   * Gets a template package from HotDocs.
   *
   * @param string $access_token
   *   The access token.
   * @param string $template_package_id
   *   The template package ID.
   *
   * @return array
   *   The template package.
   */
  protected function getTemplatePackage($access_token, $template_package_id) {
    if (empty($access_token) || empty($template_package_id)) {
      return NULL;
    }

    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.template_package',
      ['template_package_id' => $template_package_id],
      [
        'absolute' => TRUE,
        'query' => ['access_token' => $access_token],
      ]
    );

    $client = \Drupal::httpClient();
    $response = $client->get($url);
    $content = $response->getBody();
    $template_package = json_decode($content, TRUE);
    return $template_package;
  }

  /**
   * Gets a template package's variable data from HotDocs.
   *
   * @param string $access_token
   *   The access token.
   * @param string $template_package_version_id
   *   The template package version ID.
   *
   * @return array
   *   The template package variable data.
   */
  protected function getTemplatePackageVariableData($access_token, $template_package_version_id) {
    if (empty($access_token) || empty($template_package_version_id)) {
      return NULL;
    }

    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.template_package_variable_data',
      ['template_package_version_id' => $template_package_version_id],
      [
        'absolute' => TRUE,
        'query' => ['access_token' => $access_token],
      ]
    );

    $client = \Drupal::httpClient();
    $response = $client->get($url);
    $content = $response->getBody();
    $variable_data = json_decode($content, TRUE);
    return $variable_data;
  }

  /**
   * AJAX callback for template package ID field.
   */
  public function templatePackageIdAjaxCallback(array &$form, FormStateInterface $form_state) {
    $template_package_id = $form_state->getValue('template_package_id') ??
      $form_state->getValue(['settings', 'template_package_id']);

    $element = [
      '#type' => 'details',
      '#title' => $this->t('Field mappings'),
      '#description' => $this->t('Map Webform fields to HotDocs fields.'),
      '#open' => TRUE,
      '#attributes' => [
        'id' => 'edit-field-mappings',
      ],
    ];

    if (empty($template_package_id)) {
      return $element;
    }

    try {
      $access_token = $this->getAccessToken();
      if ($access_token) {
        $template_package = $this->getTemplatePackage($access_token, $template_package_id);
        $version_id = $template_package['liveVersionId'];
        $variableData = $this->getTemplatePackageVariableData($access_token, $version_id);
        $variables = $variableData['variables'];

        $element['message'] = [
          '#type' => 'html_tag',
          '#tag' => 'div',
          '#value' => $this->t('Successfully connected to HotDocs. Save the handler, then edit it to map fields.'),
          '#attributes' => [
            'class' => [
              'messages',
              'messages--status',
            ],
          ],
        ];

        $element['variables'] = [
          '#type' => 'details',
          '#title' => $this->t('Document variables'),
          '#description' => $this->t('These are the document variables from HotDocs for the selected template package.'),
          '#open' => FALSE,
          'table' => [
            '#type' => 'table',
            '#header' => [
              $this->t('Name'),
              $this->t('Type'),
            ],
            '#rows' => array_map(function ($variable) {
              return [
                $variable['name'],
                $variable['type'],
              ];
            }, $variables),
          ],
        ];
      }
    }
    catch (\Exception $e) {
      \Drupal::logger('webform_hotdocs')->error('Failed to authenticate with HotDocs: @error', ['@error' => $e->getMessage()]);

      $element['message'] = [
        '#type' => 'html_tag',
        '#tag' => 'div',
        '#value' => $this->t('Failed to authenticate with HotDocs: @error', ['@error' => $e->getMessage()]),
        '#attributes' => [
          'class' => [
            'messages',
            'messages--error',
          ],
        ],
      ];
    }

    return $element;
  }

  /**
   * Creates a work item in HotDocs.
   *
   * @param string $access_token
   *   The access token.
   * @param string $answer_set
   *   The answer set.
   * @param string $template_package_id
   *   The template package ID.
   * @param string $guid
   *   The GUID.
   *
   * @return bool
   *   TRUE if successful, FALSE otherwise.
   */
  protected function createWorkItem($access_token, $answer_set, $template_package_id, $guid) {
    if (empty($access_token) || empty($answer_set) || empty($template_package_id)) {
      return FALSE;
    }

    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.create_work_item',
      [
        'guid' => $guid,
      ],
      [
        'absolute' => TRUE,
        'query' => ['access_token' => $access_token],
      ]
    );

    // Load the workgroup ID from Webform third-party settings.
    $third_party_settings_manager = \Drupal::service('webform.third_party_settings_manager');
    $settings = $third_party_settings_manager->getThirdPartySettings('hotdocs');
    $workgroup_id = $settings['workgroup_id'];

    // Load the webform submission.
    $webform_submission = $this->getWebformSubmission();

    $body_data = [
      'name' => 'Webform submission ' . $webform_submission->id() . ' (' . $guid . ')',
      'workGroupId' => $workgroup_id,
      'templatePackageIds' => [$template_package_id],
      'answerXml' => $answer_set,
      'isPrivateToOwner' => TRUE,
    ];
    $client = \Drupal::httpClient();
    $response = $client->put($url, [
      'headers' => [
        'Content-Type' => 'application/json',
      ],
      'json' => $body_data,
    ]);

    // Return TRUE if status code is 200. The response contains no content.
    return $response->getStatusCode() === 200;
  }

  /**
   * Creates and assembles a new version of a work item in HotDocs.
   */
  protected function createAndAssembleWorkItem($access_token, $template_package_id, $work_item_id) {
    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.create_and_assemble_work_item',
      [
        'template_package_id' => $template_package_id,
        'work_item_id' => $work_item_id,
      ],
      [
        'absolute' => TRUE,
        'query' => [
          'access_token' => $access_token,
        ],
      ]
    );

    $client = \Drupal::httpClient();
    $response = $client->put($url);
    return json_decode($response->getBody()->getContents(), TRUE);
  }

  /**
   * Downloads assembled documents from HotDocs.
   */
  protected function downloadAssembledDocuments($access_token, $work_item_id) {
    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.download_assembled_documents',
      [
        'work_item_id' => $work_item_id,
      ],
      [
        'absolute' => TRUE,
        'query' => [
          'access_token' => $access_token,
        ],
      ]
    );

    $client = \Drupal::httpClient();
    $response = $client->get($url);
    $content = $response->getBody();
    $documents = json_decode($content, TRUE);
    return $documents;
  }

  /**
   * Downloads the content of an assembled document from HotDocs.
   */
  protected function downloadAssembledDocumentContent($access_token, $template_package_id, $work_item_id, $document_id) {
    $url = \Drupal::urlGenerator()->generateFromRoute('webform_hotdocs.download_assembled_document_content',
      [
        'template_package_id' => $template_package_id,
        'work_item_id' => $work_item_id,
        'document_id' => $document_id,
      ],
      [
        'absolute' => TRUE,
        'query' => ['access_token' => $access_token],
      ]
    );

    $client = \Drupal::httpClient();
    $response = $client->get($url);
    return $response->getBody()->getContents();
  }

  /**
   * Helper function to check if an element should have a mapping field.
   */
  private function isValidFormElement($element) {
    $elements_to_skip = [
      'markup',
      'webform_section',
      'fieldset',
      'webform_wizard_page',
      'processed_text',
    ];
    return isset($element['#type']) && !in_array($element['#type'], $elements_to_skip);
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE) {
    try {
      // Step 1: Get access token.
      try {
        $access_token = $this->getAccessToken();
        if (empty($access_token)) {
          throw new \Exception('Failed to obtain access token.');
        }
      }
      catch (\Exception $e) {
        \Drupal::logger('webform_hotdocs')->error('Failed to get access token: @error', ['@error' => $e->getMessage()]);
        throw $e;
      }

      $template_package_id = $this->configuration['template_package_id'];
      $guid = \Drupal::service('uuid')->generate();

      // Step 2: Prepare submission data.
      $state = $webform_submission->getWebform()->getSetting('results_disabled') ?
        WebformSubmissionInterface::STATE_COMPLETED :
        $webform_submission->getState();

      $data = $webform_submission->getData();
      $field_mappings = $this->configuration['field_mappings'];
      $fields_to_send = [];
      foreach ($field_mappings as $webform_field_name => $hotdocs_field_name) {
        if (isset($data[$webform_field_name]) && !empty($hotdocs_field_name)) {
          $fields_to_send[$webform_field_name] = [
            'field_name' => $hotdocs_field_name,
            'value' => $data[$webform_field_name],
          ];
        }
      }

      // Step 3: Create work item.
      try {
        $answer_set = $this->prepareAnswerSet($fields_to_send);

        // Log the answer set for debugging.
        \Drupal::logger('webform_hotdocs')->debug('Prepared answer set: @answer_set', [
          '@answer_set' => $answer_set,
        ]);

        $created_work_item = $this->createWorkItem($access_token, $answer_set, $template_package_id, $guid);
        if (!$created_work_item) {
          throw new \Exception('Failed to create work item.');
        }
      }
      catch (\Exception $e) {
        // Log detailed error information.
        $error_message = $e->getMessage();
        $error_trace = $e->getTraceAsString();

        \Drupal::logger('webform_hotdocs')->error('Failed to create work item. Error: @error. Trace: @trace', [
          '@error' => $error_message,
          '@trace' => $error_trace,
        ]);

        // If it's a Guzzle exception, log the request and response details.
        if ($e instanceof RequestException) {
          $response = $e->hasResponse() ? $e->getResponse() : NULL;

          if ($response) {
            \Drupal::logger('webform_hotdocs')->error('Response details: Status: @status, Body: @body', [
              '@status' => $response->getStatusCode(),
              '@body' => (string) $response->getBody(),
            ]);
          }
        }

        throw $e;
      }

      // Step 4: Assemble work item.
      try {
        $assembled_work_item = $this->createAndAssembleWorkItem($access_token, $template_package_id, $guid);
        if (empty($assembled_work_item)) {
          throw new \Exception('Failed to assemble work item.');
        }
      }
      catch (\Exception $e) {
        \Drupal::logger('webform_hotdocs')->error('Failed to assemble work item: @error', ['@error' => $e->getMessage()]);
        throw $e;
      }

      // Step 5: Download documents.
      try {
        $downloaded_documents = $this->downloadAssembledDocuments($access_token, $guid);
        $documents_data = json_decode($downloaded_documents, TRUE);
        if (empty($documents_data['items'][0]['documentId'])) {
          throw new \Exception('No document ID found in response.');
        }
        $document_id = $documents_data['items'][0]['documentId'];
        $document_type = $documents_data['items'][0]['type'];
        $document_name = $documents_data['items'][0]['name'];
      }
      catch (\Exception $e) {
        \Drupal::logger('webform_hotdocs')->error('Failed to download assembled documents: @error', ['@error' => $e->getMessage()]);
        throw $e;
      }

      // Step 6: Download document content.
      try {
        $document_content = $this->downloadAssembledDocumentContent($access_token, $template_package_id, $guid, $document_id);
        if (empty($document_content)) {
          throw new \Exception('Empty document content received.');
        }

        // Generate unique ID and store document.
        $document_id = \Drupal::service('uuid')->generate();
        $base64_content = base64_encode($document_content);
        \Drupal::keyValue('webform_hotdocs_documents')->set($document_id, [
          'content' => $base64_content,
          'type' => $document_type,
          'name' => $document_name,
        ]);

        // Store token data.
        $token_data = [
          'webform_handler' => [
            $this->getHandlerId() => [
              $state => [
                'document_id' => $document_id,
              ],
            ],
          ],
        ];

        // Update submission if needed.
        $submission_data = $webform_submission->getData();
        $submission_has_token = (strpos(print_r($submission_data, TRUE), '[webform:handler:' . $this->getHandlerId() . ':') !== FALSE) ? TRUE : FALSE;
        if ($submission_has_token) {
          $submission_data = $this->replaceTokens($submission_data, $webform_submission, $token_data);
          $webform_submission->setData($submission_data);
          if ($this->isResultsEnabled()) {
            $webform_submission->resave();
          }
        }
      }
      catch (\Exception $e) {
        \Drupal::logger('webform_hotdocs')->error('Failed to process document content: @error', ['@error' => $e->getMessage()]);
        throw $e;
      }
    }
    catch (\Exception $e) {
      \Drupal::logger('webform_hotdocs')->error('HotDocs processing failed: @error', ['@error' => $e->getMessage()]);
    }
  }

  /**
   * Helper function to determine if saving of results is enabled.
   */
  protected function isResultsEnabled() {
    return ($this->getWebform()->getSetting('results_disabled') === FALSE);
  }

  /**
   * Prepares an answer set for HotDocs.
   */
  private function prepareAnswerSet($fields) {
    if (empty($fields)) {
      return NULL;
    }

    $hotdocs_variables = $this->configuration['hotdocs_variables'];

    $answer_set = '<hd:answerSet title="" version="2.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:hd="http://www.hotdocs.com/schemas/answer_set/2017">';
    foreach ($fields as $field) {
      if (!empty($field['value'])) {
        $hotdocs_variable = $hotdocs_variables[$field['field_name']];
        $answer_set .= $this->prepareAnswer($field['field_name'], $field['value'], $hotdocs_variable);
      }
    }
    $answer_set .= '</hd:answerSet>';

    // Validate the answer set.
    if (!$this->validateAnswerSet($answer_set)) {
      throw new \Exception('The answer set is not valid.');
    }

    return $answer_set;
  }

  /**
   * Prepares an answer based on the field type.
   */
  private function prepareAnswer($name, $value, $hotdocs_variable) {
    switch ($hotdocs_variable['type']) {
      case 'textVarType':
      case 'singleSelectVarType':
        // If the value is an array (example: Name), implode it with a space.
        if (is_array($value)) {
          $value = implode(' ', array_filter($value));
        }
        return '<hd:textAnswer name="' . $name . '"><hd:value>' . htmlspecialchars($value) . '</hd:value></hd:textAnswer>';

      case 'numberVarType':
        return '<hd:numberAnswer name="' . $name . '"><hd:value>' . floatval($value) . '</hd:value></hd:numberAnswer>';

      case 'dateVarType':
        $date = new \DateTime($value);
        return '<hd:dateAnswer name="' . $name . '"><hd:value>' . $date->format('Y-m-d') . '</hd:value></hd:dateAnswer>';

      case 'trueFalseVarType':
        return '<hd:trueFalseAnswer name="' . $name . '"><hd:value>' . ($value ? 'true' : 'false') . '</hd:value></hd:trueFalseAnswer>';

      case 'multipleChoiceVarType':
        $selections = is_array($value) ? $value : [$value];
        return '<hd:recordListAnswer name="' . $name . '"><hd:values><hd:value>' . implode('', array_map(function ($selection) {
            return '<hd:text name="' . htmlspecialchars($selection) . '"></hd:text>';
        }, $selections)) . '</hd:value></hd:values></hd:recordListAnswer>';

      default:
        return '';
    }
  }

  /**
   * Validates an answer set against the HotDocs schema.
   */
  private function validateAnswerSet($answer_set) {
    libxml_use_internal_errors(TRUE);

    $xml = new \DOMDocument();
    $xml->loadXML($answer_set);

    // Get the schema path from within the module.
    $schema = \Drupal::service('extension.list.module')->getPath('webform_hotdocs') . '/schemas/AnswerSet_2017.xsd';

    if (!$xml->schemaValidate($schema)) {
      $errors = libxml_get_errors();
      libxml_clear_errors();

      // Log validation errors.
      foreach ($errors as $error) {
        \Drupal::logger('webform_hotdocs')->error('Answer Set validation error: @error', ['@error' => $error->message]);
      }

      return FALSE;
    }

    return TRUE;
  }

}
